/*
	glitterclock.c

	code by SASANO Takayoshi
	--- public domain, no warranty.
*/

#include <driver/usb.h>
#include <btron/clk.h>
#include <bstdio.h>
#include <errcode.h>

#define	DEFAULTPRI	80
#define	STACKSIZE	2048

#ifdef	DEBUG
#define	DP(x)	printf x
#else
#define	DP(x)	/* */
#endif

LOCAL	ID	PorID;
LOCAL	ID	TskID;
LOCAL	W	usbIfID = -1;
LOCAL	W	usbEpID = -1;
LOCAL	W	Tick = -1;

#define	CMD_SIZE	6
#define	DAT_SIZE	32
#define	BUF_SIZE	(CMD_SIZE + DAT_SIZE)

/* žѥХåե */
LOCAL	UB	XferBuf[BUF_SIZE] = {
	0x80,	// ѥޥɡ6byte, ܺ٤
	0x03,
	0x58,	// bit[3:2] brightness 11:100% 10:75% 01:50% 00:25%
	0x08,
	0x00,
	0x00,
		// Ĥѥǡʥꥢѡ
};

/* եȥǡ */
LOCAL	UB	digit4x8[][8] = {
	{0xe, 0xa, 0xa, 0xa, 0xa, 0xa, 0xe, 0x0},	// 0
	{0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0},	// 1
	{0xe, 0x2, 0x2, 0xe, 0x8, 0x8, 0xe, 0x0},	// 2
	{0xe, 0x2, 0x2, 0xe, 0x2, 0x2, 0xe, 0x0},	// 3
	{0xa, 0xa, 0xa, 0xe, 0x2, 0x2, 0x2, 0x0},	// 4
	{0xe, 0x8, 0x8, 0xe, 0x2, 0x2, 0xe, 0x0},	// 5
	{0xe, 0x8, 0x8, 0xe, 0xa, 0xa, 0xe, 0x0},	// 6
	{0xe, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0},	// 7
	{0xe, 0xa, 0xa, 0xe, 0xa, 0xa, 0xe, 0x0},	// 8
	{0xe, 0xa, 0xa, 0xe, 0x2, 0x2, 0xe, 0x0},	// 9
	{0x0, 0x0, 0x2, 0x0, 0x2, 0x0, 0x0, 0x0},	// :
	{0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0},	// (space)
};

/* ɽǡ */
LOCAL	void	create_data(void)
{
	int	i, h;
	STIME	t;
	DATE_TIM	tm;
	UB	*hr10, *hr1, *colon, *min10, *min1;
	UB	*buf = &XferBuf[CMD_SIZE + 8];

	b_get_tim(&t, NULL);
	b_get_tod(&tm, t, True);

	h = tm.d_hour % 12;
	if (h == 0) h = 12;

	hr10  = digit4x8[(h >= 10) ? 1 : 11];
	hr1   = digit4x8[h % 10];
	colon = digit4x8[(Tick++ & 2) ? 10 : 11];	// äƱ
	min10 = digit4x8[tm.d_min / 10];
	min1  = digit4x8[tm.d_min % 10];

	for (i = 0; i < 8; i++) {
		buf[i * 2 + 0] = (*hr10++ << 6) | (*hr1++ << 2) | *colon++;
		buf[i * 2 + 1] = (*min10++ << 4) | *min1++;
	}

	return;
}

/* ؼ */
LOCAL	void	redraw(void)
{
	W	er;

	/* ɽǡκ */
	create_data();

	/* åѥͥ */
	er = usbIoPipe(usbEpID, XferBuf, sizeof(XferBuf), NULL);
	if (er < USB_OK) {
		DP(("redraw: usbIoPipe %d\n", er));
		goto fin0;
	}

fin0:
	return;
}

/* USB٥Ƚ */
LOCAL	W	usb_event(DevReq *q, DevRsp *r)
{
#define	VENDOR_ID	0x0f30
#define	PRODUCT_ID	0x0040
#define	EP_ADR_OUT	0x02

	W	er, did, iid, pid, ifno;
	usbDeviceDescriptor	d;

	/* ǥХ׵ */
	if (q->datano == USB_UNMOUNT) {
		DP(("usb_event: detach\n"));
		iid = usbIfID;
		pid = usbEpID;
		usbIfID = usbEpID = Tick = -1;
		er = USB_OK;
		goto fin2;
	}

	did = q->devid & 0x7f;
	ifno = (q->devid >> 8) & 0xff;

	/* ʤΥå */
	if (usbDescriptorDevice(did, (UB *)&d, sizeof(d), NULL) < USB_OK ||
	    d.idVendor != VENDOR_ID || d.idProduct != PRODUCT_ID) {
		DP(("usb_event: not target\n"));
		er = USB_ERR_DEVICE;
		goto fin0;
	}

	/* 󥿥եΥץ */
	er = usbOpenInterface(did, ifno);
	if (er < USB_OK) {
		DP(("usb_event: usbOpenInterface %d\n", er));
		goto fin0;
	}
	iid = er;

	/* ɥݥȤΥץ */
	er = usbOpenPipe(iid, EP_ADR_OUT, USB_WAIT, -1);
	if (er < USB_OK) {
		DP(("usb_event: usbOpenPipe %d\n", er));
		goto fin1;
	}
	pid = er;

	/* ɥݥȤν */
	er = usbStallPipe(pid, 2);
	if (er < USB_OK) {
		DP(("usb_event: usbOpenPipe %d\n", er));
		goto fin2;
	}

	DP(("usb_event: attached\n"));
	usbIfID = iid;
	usbEpID = pid;

	/* ³餹ɽ褦׵Ф */
	redraw();

	er = USB_OK;
	goto fin0;

fin2:
	usbClosePipe(pid);
fin1:
	usbCloseInterface(iid);
fin0:
	return (er < USB_OK) ? USB_NONE : USB_OWN;
}

/*  */
LOCAL	void	io_maintask(void)
{
#define	TIMEOUT	250

	W	er, size;
	RNO	rno;
	DevReq	q;
	DevRsp	r;

	while (1) {
		/* ׵Ԥ */
		er = tacp_por(&rno, (VP)&q, &size, PorID,
			      D_CALPTN(DC_USBEVENT), TIMEOUT);

		/* ॢȥ顼Ѥ׵Ф */
		// ɽPCλ狼TIMEOUT(msec)٤
		if (er < E_OK) {
			if (er == E_TMOUT) {
				if (usbIfID >= 0) redraw();
				continue;
			} else {
				goto fin0;
			}
		}

		/* ʥΥѥåȤ̵ */
		if (size < sizeof(DevReq)) {
			DP(("io_maintask: bad packet size %d\n", size));
			continue;
		}

		/* ѥѥåȤν׵μ¹ */
		r.devid     = q.devid;
		r.cmd       = q.cmd;
		r.datano    = q.datano;
		r.datacnt   = 0;
		r.error.err = (q.cmd.cmd == DC_USBEVENT) ?
			usb_event(&q, &r) : ER_NOSPT;

		/* ǥ֤ؤ */
		er = rpl_rdv(rno, (VP)&r, sizeof(DevRsp));
		if (er < E_OK) {
			DP(("io_maintask: rpl_por %d\n", er));
		}
	}

fin0:
	exd_tsk();
}

/* ɥ饤Сȥݥ */
EXPORT	W	main(Bool start, TC *arg)
{
	W	er;
	T_CPOR	cpor = {
		.exinf = NULL,
		.poratr = TA_NULL,
		.maxcmsz = sizeof(DevReq),
		.maxrmsz = sizeof(DevRsp),
	};
	T_CTSK	ctsk = {
		.exinf = NULL,
		.task = io_maintask,
		.itskpri = DEFAULTPRI,
		.stksz = STACKSIZE,
		.tskatr  = TA_HLNG | TA_RNG0,
	};
	usbEventPattern	evp = {
		.bClass = 0xff,
		.bSubClass = 0xff,
		.bProtocol = 0xff,
		.mask = EVENT_CLASS | EVENT_SUBCLASS | EVENT_PROTOCOL,
	};

	/* λ׵ */
	if (!start) {
		if (usbIfID >= 0) usbCloseInterface(usbIfID);
		er = E_OK;
		goto fin4;
	}

	/* ǥ֥ݡȺ */
	er = vcre_por(&cpor);
	if (er < E_OK) {
		DP(("main: vcre_por %d\n", er));
		goto fin0;
	}
	PorID = (ID)er;

	/* դ */
	er = vcre_tsk(&ctsk);
	if (er < E_OK) {
		DP(("main: vcre_tsk %d\n", er));
		goto fin1;
	}
	TskID = (ID)er;

	/* դư */
	er = sta_tsk(TskID, 0);
	if (er < E_OK) {
		DP(("main: sta_tsk %d\n", er));
		goto fin2;
	}

	/* USB󥿥եɥ饤ФȤ */
	er = usbRegistInterface(PorID, &evp);
	if (er < USB_OK) {
		DP(("main: usbRegistInterface %d\n", er));
		goto fin3;
	}

	er = ER_OK;
	goto fin0;

fin4:
	usbRegistInterface(PorID, NULL);
fin3:
	ter_tsk(TskID);
fin2:
	del_tsk(TskID);
fin1:
	del_por(PorID);
fin0:
	return er;
}
